import snappy
import hikmot
import math
from math import sqrt, pi
from imath import isqrt
import hikmot.interval as itl
import hikmot.complex as ic
from datetime import datetime
import cmath
#from cmath import log, exp
import copy

# constants.py contains the global constants used in all files

from constants import *

# functions.py contains the basic numerical functions used here

from functions import *

# exclude.py contains the list of slopes that are excluded from the analysis

from exclude import *

# manifold.py contains functions that use the manifold class

from manifold import *

# cusp.py contains functions that deal with cusp (shape and neighborhood)

from cusp import *

# A 'slope' is a pair (p,q) of coprime integers with q>=0, 
# with (-1,0) excluded (because (1,0) suffices). 
# Slopes are in 1-1 correspondence with esssential simple curves on a torus, 
# after fixing a homology basis (as usual).
# The 'empty slope' is denoted by (0, 0)

# A 'filling' is a sequence of slopes, that assigns a slope to each cusp of the manifold. 
# The empty slope (0, 0) denotes an unfilled cusp, as in Snappy.
# For instance, [(-1,1), (0,0), (7,3)] is a filling on a 3-cusped manifold.

# Here you must set the manifold N you want to investigate. 
# Some examples are listed below.

# N = snappy.Manifold('4_1')             	# The figure-8 knot complement
# N = snappy.Manifold('Whitehead.tri')   	# The Whitehead link complement
# N = snappy.Manifold('chain3.tri')      	# The magic manifold (chain link with 3 components)
# N = snappy.Manifold('chain4.tri')      	# The chain link with 4 components with vol = 2octahedra
# N = snappy.Manifold('chain5.tri')      	# The chain link with 5 components
# N = snappy.Manifold('chain5_fill_m2.tri')   # The chain link with 5 components filled with -2
# N = snappy.Manifold('chain6.tri')      	# The chain link with 5 components

N = snappy.Manifold('manifold.tri')

# This is the main routine, which finds all exceptional slopes of N.
# The 'tetrahedra_shapes' are the shapes of N, priorly obtained using hikmot.verify_hyperbolicity(N)
# 'filling' is the first filling which is explored: it is used as an internal counter,
#	since the function calls itself very often. To start, set it as [(0, 0), ... (0, 0)]
# The routine restricts the search to fillings that have exactly 'num_complete_required' complete cusps.
# The resulting exceptional fillings are appended to the list 'exceptional_fillings'.
#	Those having low degeneracy (<=1) are also appended to the list 'maybe hyperbolic'.
# The search avoids fillings that contain some elements in the lists 'avoid', 'exclude', 'already_seen'
# 	(the list 'already_seen' is for internal use, to speed up the algorithm).

def find_exceptional_fillings(N, tetrahedra_shapes, filling, exceptional_fillings, num_complete_required, avoid, already_seen, maybe_hyperbolic):
	global max_A_diff
	num_cusps = len(filling)
	num_complete = filling.count((0,0))
	if num_complete <= num_complete_required:
		return 

	# Set the area A of the maximal cusp section (with all cusps tied)

	A = 0
	if use_SnapPea_functions == True: 
		A = maximal_area(N)   
	if use_our_functions == True:
		A_SnapPea = itl.interval(A)
		A, canonized_N, new_shapes = get_maximal_cusp_section(N)

		# To calculate A we used canonization and hikmot. 
		# If either fails, the functions return A = -1.

		if A == -1:

			# We set A = (3 * sqrt(3))/8 = 0.649519052838...
			# Cusp sections with that area are guaranteed to exist.

			A = (3 * isqrt(3))/8
			if verbose >= 2:
				print "failed to find a cusp section"

		else:

			# Sanity check: compare A_Snappea and our A

			if A_SnapPea.inf > 0 and A.inf > 0:
				A_diff = A_SnapPea - A	

				# if A_diff is negative there is probably something wrong with our functions 

				if verbose and A_diff.sup < 0:
					print 'ACHTUNG. Our value for cusp area A is bigger than the one calculated by SnapPea.'
					print "A_diff = ", A_diff.sup, 'A_SnapPea =', A_SnapPea.sup, 'A_us =', A.sup, 
				if A_diff > max_A_diff:
					max_A_diff = A_diff
				if verbose == 2:
					print 'A_diff, max_A_diff =', A_diff, max_A_diff, A

	# The cusps in N form a subset of those in 'filling' because some non-complete
	# cusp in N might have been permanently filled at some previous stage. 
	# We thus unfortunately need two counters here: real_cusp and original_cusp;
	# they will denote the same complete cusp.

	real_cusp = -1		
	original_cusp = -1
	for cusp in range (0, num_complete):
		real_cusp = real_cusp + 1
		while N.cusp_info('complete?')[real_cusp] == False:
			real_cusp = real_cusp + 1
		original_cusp = original_cusp + 1
		while filling[original_cusp] != (0,0):
			original_cusp = original_cusp + 1

		# Encode the cusp shape x + iy.

		(x, y) = (0, 0)
		if use_SnapPea_functions == True: 
			(x, y) = cusp_shape(N.filled_triangulation(), real_cusp)
		if use_our_functions == True: 
			(x_SnapPea, y_SnapPea) = (x, y)
			(x, y) = get_cusp_shape(N, real_cusp, tetrahedra_shapes)

			# If the difference between our and Snappy's cusp shape is bigger than
			# some error, then print it

			if (x_SnapPea, y_SnapPea) != (0, 0):
				error = 0.001
				if abs(x - x_SnapPea) > error or abs(y - y_SnapPea) > error:
					print 'ACHTUNG. The difference between the cusp shapes measured by \
						the two functions is bigger than', error

		# List all short slopes:

		short_slopes = get_short_slopes(A, x, y)
		if verbose >= 2:
			print "Slopes with length < 6.001:", short_slopes;

		# For each short slope, fill the manifold. Then iterate.
		
		for slope in short_slopes:

			# Encode the filled slope in the array 'filling'

			filling[original_cusp] = slope

			# If the resulting 'filling' contains elements of 'avoid', 'already_seen', or 'exclude',
			# then forget about it.
			
			avoid_it = False
			if list_contains(filling, avoid) or list_contains(filling, already_seen) or list_contains(filling, exclude):
				avoid_it = True

			if avoid_it == False:

				if verbose: 
					print "---", filling

				# Fill the manifold. We work on a copy.
				
				M = N.copy()	
				M.dehn_fill(slope, real_cusp)
				M.simplify()

				# Try to hyperbolize M. The resulting M2 is the triangulation with
				# lowest degeneracy among those found by the routine.
				# Note that M2 might or not be permanently filled

				flag, M2, M2_shapes = try_to_hyperbolize(M, number_of_randomizes, try_permanent_fillings, min_imaginary) 

				# If it is hyperbolic, we are happy. If there are some complete cusps, iterate.
				# 'Hyperbolic' for us means positive solution ( > min_imaginary) and hikmot verified

				if flag:
					if verbose >= 2:
						print M2, 'Volume = ', M2.volume()
						print M2.solution_type(), 'minimum_imaginary:', min_imaginary_shapes(M2)
					if (num_complete > num_complete_required + 1):
						find_exceptional_fillings(M2, M2_shapes, filling, exceptional_fillings, num_complete_required, avoid, already_seen, maybe_hyperbolic)

				# If it is not hyperbolic, add filling to the 'exceptional_fillings' list.

				else:
					exceptional_fillings.append(filling[:])  

					# However, if the triangulation has a triangulation with flat or negatively oriented tetrahedra,
					# we suspect that it is hyperbolic. So we add it to the 'maybe_hyperbolic' list.

					if degeneracy_level[M2.solution_type()] <= 1:
						maybe_hyperbolic[num_cusps - num_complete_required -1].append(filling[:])
						if verbose:
							print 'might be hyperbolic:'
							print M2.solution_type(), 'minimum_imaginary:', min_imaginary_shapes(M2)

				# Append 'filling' to 'already_seen', so that if we encounter it again we can skip it.

				already_seen.append(filling[:])

		# Set back the cusp to the empty filling.

		filling[original_cusp] = (0,0) 
	return

# Applies an isometry to a filling

def apply_isometry(isometry, filling):
	num_cusps = len(filling)
	new_filling = filling[:]
	for i in range(num_cusps):
		matrix = isometry.cusp_maps()[i].data
		slope = filling[i]
		mapped_slope = (matrix[0][0] * slope[0] + matrix[0][1] * slope[1], matrix[1][0] * slope[0] + matrix[1][1] * slope [1])

		# Recall that a slope here is a pair (p,q) with q>=0 and is not (-1,0):

		if mapped_slope[1]<0:
			mapped_slope = (-mapped_slope[0], -mapped_slope[1])
		if mapped_slope == (-1,0):
			mapped_slope = (1,0)
		new_filling[isometry.cusp_images()[i]] = mapped_slope
	return new_filling



# The main program starts here

time1 = datetime.now()

print 'Manifold volume:'
print N.volume()
print 'Cusp shapes:'
print N.cusp_info()

flag, tetrahedra_shapes = hikmot.verify_hyperbolicity(N, verbose >=2)
if flag == False:
	print 'Cannot verify hyperbolicity of the original manifold.'
	exit

G = N.symmetry_group()
isometries = G.isometries()
if verbose:
	print G
		
num_cusps = N.num_cusps()

# Exceptional_fillings[i] will be the list of all isolated exceptional fillings of N
# having i+1 filled cusps. 
# (An exceptional filling is isolated if it does not contain a proper exceptional filling)

exceptional_fillings = []

# Exceptional_fillings_quotient[i] will be the quotient of exceptional_fillings[i]
# up to the action of the isometry group of N

exceptional_fillings_quotient = []

# Maybe_hyperbolic[i] is a subset of exceptional_fillings[i] consisting of those manifolds
# for which we could find some solutions with degeneracy <=1 (negatively oriented tetrahedra)

maybe_hyperbolic = []

# Filling is a counter used by the routine 'find_exceptional_fillings'. 
# It is settled as [(0,0), ..., (0,0)]

filling = []
for i in range(num_cusps):
	filling.append ((0,0))
	exceptional_fillings.append ([])
	maybe_hyperbolic.append([])
	exceptional_fillings_quotient.append([])

# Avoid is a list which will be used by 'find_exceptional_fillings' to avoid
# investigating some fillings that were already identified as exceptional.

avoid = []

# Classify the exceptional fillings of the hyperbolic manifold N

for i in range(num_cusps):	
	if verbose:
		print 'With', i+1, 'fillings:'
	find_exceptional_fillings(N, tetrahedra_shapes, filling, exceptional_fillings[i], num_cusps-i-1, avoid, [], maybe_hyperbolic)

	# Next time, the algorithm must avoid the fillings containing those that we have
	# just found to be exceptional. We store this information in 'avoid'.

	avoid.extend(exceptional_fillings[i][:])

	# Reset fillings.

	for j in range(num_cusps):
		filling[j] = (0,0)

# Sort the results, eliminate maybe_hyperbolic from exceptional_fillings

for i in range(num_cusps):   

	# Remove maybe_hyperbolic from exceptional_fillings

	for filling in maybe_hyperbolic[i]:
		exceptional_fillings[i].remove(filling)
	exceptional_fillings[i].sort()
	maybe_hyperbolic[i].sort()

# If 'quotient_by_isometries' is True,
# create quotient list of candidate exceptional fillings up to the action of the isometry group:

if quotient_by_isometries == True:
	for i in range(num_cusps):
		for filling in exceptional_fillings[i]:

			# Among all images of filling through isometries, we take the lexicographically smallest one:
			# This gives a unique representative on each equivalence class of fillings.

			new_filling = filling[:]
			for isometry in isometries:
				image_filling = apply_isometry(isometry, filling)
				if image_filling < new_filling:
					new_filling = image_filling 
			exceptional_fillings_quotient[i].append(new_filling[:])
		exceptional_fillings_quotient[i].sort()
		eliminate_duplicates(exceptional_fillings_quotient[i])

# Print candidate exceptional_fillings

print 'Candidate exceptional fillings:'
for i in range(num_cusps):
	print 'With', i+1, 'fillings:'
	print exceptional_fillings[i]
	print 'Total:', len(exceptional_fillings[i])

# If quotient_by_isometries is True,
# print the quotient of the previous list up to the action of the isometry group:

if quotient_by_isometries == True:
	print 'Candidate exceptional fillings, up to the action of the isometry group:'
	for i in range(num_cusps):
		print 'With', i+1, 'fillings:'
		print exceptional_fillings_quotient[i]
		print 'Total:', len(exceptional_fillings_quotient[i])

# Print candidate hyperbolic fillings

print 'Candidate hyperbolic fillings:'
for i in range(num_cusps):
	print 'With', i+1, 'fillings:'
	print maybe_hyperbolic[i]
	print 'Total:', len(maybe_hyperbolic[i])

if verbose and use_our_functions and use_SnapPea_functions:
	print 'Maximum discrepancy between our area and SnapPea''s area', max_A_diff

time2 = datetime.now()
if verbose:
	print 'Computer time needed:', time2 - time1