# This file contains the basic functions used by find_exceptional_fillings.py

from imath import isqrt
import hikmot.interval as itl

# Eliminates duplicates from a list

def eliminate_duplicates(list):
	l = len(list)
	i = 0
	while i<l-1:
		if list[i] == list[i+1]:
			del list[i]
			l = l-1
		else:
			i = i+1

# Returns True iff the pair of integers (a,b) determines a slope 
# ie if they are coprime and b>=0 and different from (-1,0)

def real_slope(a, b):
	if (a==0 and b<>1) or (b==0 and a<>1): return False
	if b<0: return False
	i = 2
	if a<0: a=-a
	while i <= a and i <= b:
		if a%i == 0 and b%i == 0: return False
		i = i+1
	return True

# Tells whether 'filling' contains 'basic_filling', that is if 'filling'
# may be obtained from 'basic_filling' by filling some more slopes. 
# For instance, the filling [(1, 0), (-1, 1), (0, 0)] 
# contains the filling      [(0, 0), (-1, 1), (0, 0)]
# (if filling == basic_fillings it returns True)

def contains(filling, basic_filling):
	l = len(filling)
	for j in range(l):
		if (filling[j] != basic_filling[j]) and (basic_filling[j] != (0,0)):
			return False
	return True

# Tells whether 'filling' contains an element in the list 'basic_fillings'. 
# 'Contains' here has the meaning explained in the previous routine!
# For instance, the filling [(1, 1), (2,1)] contains the second element of the list
# [[(1, 0), (0, 0)], [(1, 1), (0, 0)]]

def list_contains(filling, basic_fillings):
	for basic_filling in basic_fillings:
		if contains(filling, basic_filling):
			return True
	return False

# Returns the (squared) lenght of a geodesic determined py the pair (p, q) in an euclidean torus
# with area A and shape (x + iy)

def squared_length(A, x, y, p, q):

	# The formula is (A/y) * ((p+x*q)**2 + (y*q)**2)
	a = (p + x * q)

	return (A/y) * (a * a + ((y * q) * (y * q)))

# Returns the list of all short slopes (i.e. those having length <= K = 6.1) 
# in a euclidean torus with area A and shape x + iy. 

def get_short_slopes(A, x, y):

	K = itl.interval(6.0001)
	short_slopes = []

	# If l^2(px+qy)<=K^2, then (A/y)*(yq)^2 <= K^2 which implies that q^2 <= K^2/(Ay)
	# Then qmax = K/sqrt(A*y)

	qmax = K/isqrt(A*y)   
	for q in range (0, int(qmax.sup)+1):

		# We have (p+xq)^2 + (yq)^2 <= K^2y/A and therefore (p+xq^2) <= H^2 with
		# H = sqrt((K*K*y)/A - (y*q)**2)

		H = isqrt((K * K * y)/A - (y * q) * (y * q))

		# p ranges from -H - x*q to H - x*q

		pmin = -H - x * q
		pmax = H - x * q

		# We now select among finitely many slopes (p, q) the really short ones

		for p in range (int(pmin.inf)-1, int(pmax.sup)+1):
			if real_slope(p,q)==1:

				# checks that it is indeed a short slope:

				l = squared_length(A, x, y, p, q)
				if l.inf < (K * K).sup:
					slope = (p,q)
					short_slopes.append(slope)
	return short_slopes

