# This file contains some functions that use the manifold class.

from constants import *
import hikmot

# A dictionary gives a number for each degeneracy level of solution

degeneracy_level = {'all tetrahedra positively oriented': 0,
'contains negatively oriented tetrahedra': 1,
'contains flat tetrahedra': 2,
'contains degenerate tetrahedra': 3,
'unrecognized solution type': 4,
'no solution found': 5,
'not attempted': 6}

# Returns the minimum imaginary part of all tetrahedra shapes

def min_imaginary_shapes(M):
	imaginary_shapes = []
	for i in M.tetrahedra_shapes(part = 'rect'):
		imaginary_shapes.append(i.imag)
	return min(imaginary_shapes)

# Checks that the solution is positive
# Here M.solution_type == 'all tetrahedra positively oriented' is not enough,
# we require that the imaginary part of every tetrahedron shape 
# is bigger than the constant 'min_imaginary'

def positive_solution(M, min_imaginary):
	return min_imaginary_shapes(M) > min_imaginary

# Tries to hyperbolize the manifold M. 
# It returns a copy of M whose solution has the least degree of degeneracy 
# among those investigated.
# More precisely, it returns a triple flag, manifold, tetrahedra_shapes
# where flag == True if 'manifold' is verified hyperbolic

def try_to_hyperbolize(M, number_of_randomizes, try_permanent_fillings, min_imaginary):
	answer = M.copy()

	# The best solution is: 
	# 1) 'all tetrahedra positively oriented'
	# 2) positive solution == True, which checks that imaginary parts are big enough  
	# 3) hikmot verified
	# If the algorithm finds such a best solution, it stops and returns it.

	if degeneracy_level[answer.solution_type()] == 0 and positive_solution(answer, min_imaginary):
		flag, tetrahedra_shapes = hikmot.verify_hyperbolicity(answer, verbose >=2)
		if flag:
			return True, answer, tetrahedra_shapes

	Mcopy = answer.copy()

	# Try to randomize the triangulation 'number_of_randomizes' times.

	for tentative in range(number_of_randomizes):
		Mcopy.randomize()
		if degeneracy_level[Mcopy.solution_type()] == 0 and positive_solution(Mcopy, min_imaginary):
			flag, tetrahedra_shapes = hikmot.verify_hyperbolicity(Mcopy, verbose >=2)
			if flag:
				return True, Mcopy, tetrahedra_shapes
		if degeneracy_level[Mcopy.solution_type()] < degeneracy_level[answer.solution_type()]:
			answer = Mcopy.copy()

	# Try to permanently fill some cusps (if try_permanent_fillings == True):

	if try_permanent_fillings == True: 
		num_cusps = M.num_cusps() 
		if num_cusps >= 2:		
			for i in range(num_cusps):
				if M.cusp_info('complete?')[i] == False:
					Mfilled = M.filled_triangulation([i])		
					flag, another_answer, other_shapes = try_to_hyperbolize(Mfilled, number_of_randomizes, try_permanent_fillings, min_imaginary)
					if flag == True:
						return True, another_answer, other_shapes
					if degeneracy_level[another_answer.solution_type()] < degeneracy_level[answer.solution_type()]:
						answer = another_answer.copy()
	return False, answer, []
